"""Utilities specific to Upsun deployments."""

from pathlib import Path
import sys


def get_project_name(output_str):
    """Get the project name from the output of `upsun project:info`.

    Command is run with `--format csv` flag.

    Returns:
        str: project name
    """
    lines = output_str.splitlines()
    title_line = [line for line in lines if "title," in line][0]
    # Assume first project is one to use.
    project_name = title_line.split(",")[1].strip()

    return project_name


def get_org_ids_names(output_str):
    """Get org ids and names from output of `upsun organization:list --yes --format csv`.

    Sample input:
        Name,Label,Owner email
        <org-name>,<org-label>,<org-owner@example.com>
        <org-name-2>,<org-label-2>,<org-owner-2@example.com>

    Returns:
        tuple: list, list
        None: If user has no organizations.
    """
    if "No organizations found." in output_str:
        return None

    lines = output_str.split("\n")[1:]

    org_ids = [line.split(",")[0] for line in lines if line]

    # Build descriptive names like this: "<org-id> <org-label> (<flexible|fixed>)"
    org_names = [
        f"{line.split(",")[0]} {line.split(",")[1]} ({line.split(",")[2]})"
        for line in lines
        if line
    ]

    return org_ids, org_names


def fix_git_exclude_bug():
    """Fix the bug where .upsun/local/ is not ignored on Windows.

    See: https://github.com/platformsh/cli/issues/286

    Returns:
        Str: confirmation message if bug was fixed
        None: if no changes were made
    """
    if sys.platform != "win32":
        return

    path_upsun_local = Path(".upsun") / "local"
    if not path_upsun_local.exists():
        return
    
    path_exclude = Path(".git") / "info" / "exclude"
    if not path_exclude.exists():
        return
    
    exclude_text = path_exclude.read_text()
    exclude_text_fixed = exclude_text.replace(r"/.upsun\local", r"/.upsun/local")

    if exclude_text_fixed != exclude_text:
        path_exclude.write_text(exclude_text_fixed)
        return "Fixed /.upsun/local entry in .git/info/exclude file."



"""Manages all Upsun-specific aspects of the deployment process."""

import django_simple_deploy

from dsd_upsun.platform_deployer import PlatformDeployer
from .plugin_config import PluginConfig
from . import utils as upsun_utils


@django_simple_deploy.hookimpl
def dsd_get_plugin_config():
    """Get platform-specific attributes needed by core."""
    plugin_config = PluginConfig()
    return plugin_config


@django_simple_deploy.hookimpl
def dsd_pre_inspect():
    """Do some work before core inspects the user's project."""
    # There's an apparent bug in the Upsun CLI that causes the .upsun/local/
    # dir to not be ignored on Windows like it is on other OSes. We can fix
    # that ourselves until Upsun updates their CLI.
    # In the configuration-only approach, the user has already run
    # `upsun create`, so we'll try that fix here. This avoids the user running
    # `manage.py deploy` with an unclean Git status.
    if (msg_fixed := upsun_utils.fix_git_exclude_bug()):
        return msg_fixed


@django_simple_deploy.hookimpl
def dsd_deploy():
    """Carry out platform-specific deployment steps."""
    platform_deployer = PlatformDeployer()
    platform_deployer.deploy()



{{current_settings}}

# Upsun settings.
import os

if os.environ.get("PLATFORM_APPLICATION_NAME"):
    # Import some Upsun settings from the environment.
    from platformshconfig import Config

    config = Config()

    try:
        ALLOWED_HOSTS.append("*")
    except NameError:
        ALLOWED_HOSTS = ["*"]

    DEBUG = False

    STATIC_URL = "/static/"

    if config.appDir:
        STATIC_ROOT = os.path.join(config.appDir, "static")
    if config.projectEntropy:
        SECRET_KEY = config.projectEntropy

    if not config.in_build():
        db_settings = config.credentials("database")
        DATABASES = {
            "default": {
                "ENGINE": "django.db.backends.postgresql",
                "NAME": db_settings["path"],
                "USER": db_settings["username"],
                "PASSWORD": db_settings["password"],
                "HOST": db_settings["host"],
                "PORT": db_settings["port"],
            },
            "sqlite": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
            },
        }



"""Manages all Upsun-specific aspects of the deployment process."""

import sys, os, subprocess, time
from pathlib import Path

from django.conf import settings
from django.core.management.utils import get_random_secret_key
from django.utils.crypto import get_random_string
from django.utils.safestring import mark_safe

from django_simple_deploy.management.commands.utils import plugin_utils
from django_simple_deploy.management.commands.utils.plugin_utils import dsd_config
from django_simple_deploy.management.commands.utils.command_errors import (
    DSDCommandError,
)

from . import deploy_messages as upsun_msgs
from . import utils as upsun_utils


class PlatformDeployer:
    """Perform the initial deployment to Upsun.

    If --automate-all is used, carry out an actual deployment.
    If not, do all configuration work so the user only has to commit changes, and call
    `upsun push`.
    """

    def __init__(self):
        self.templates_path = Path(__file__).parent / "templates"

    # --- Public methods ---

    def deploy(self, *args, **options):
        """Coordinate the overall configuration and deployment."""
        plugin_utils.write_output("\nConfiguring project for deployment to Upsun...")

        self._validate_platform()

        self._prep_automate_all()
        self._modify_settings()
        self._add_requirements()
        self._add_platform_app_yaml()
        self._add_platform_dir()
        self._add_services_yaml()
        self._settings_env_var()

        self._conclude_automate_all()
        self._show_success_message()

    # --- Helper methods for deploy() ---

    def _validate_platform(self):
        """Make sure the local environment and project supports deployment to
        Upsun.

        Make sure CLI is installed, and user is authenticated. Make sure necessary
        resources have been created and identified, and that we have the user's
        permission to use those resources.

        Returns:
            None

        Raises:
            DSDCommandError: If we find any reason deployment won't work.
        """
        if dsd_config.unit_testing:
            # Unit tests don't use the CLI. Use the deployed project name that was
            # passed to the simple_deploy CLI.
            self.deployed_project_name = dsd_config.deployed_project_name
            plugin_utils.log_info(
                f"Deployed project name: {self.deployed_project_name}"
            )
            return

        self._check_upsun_settings()
        self._validate_cli()

        self.deployed_project_name = self._get_upsun_project_name()
        plugin_utils.log_info(f"Deployed project name: {self.deployed_project_name}")

        self.org_name = self._get_org_name()
        plugin_utils.log_info(f"\nOrg name: {self.org_name}")

    def _prep_automate_all(self):
        """Intial work for automating entire process.

        Returns:
            None: If creation of new project was successful.

        Raises:
            DSDCommandError: If create command fails.

        Note: create command outputs project id to stdout if known, all other
          output goes to stderr.
        """
        if not dsd_config.automate_all:
            return

        plugin_utils.write_output("  Running `upsun create`...")
        plugin_utils.write_output(
            "    (Please be patient, this can take a few minutes."
        )
        cmd = f"upsun create --title { self.deployed_project_name } --org {self.org_name} --region {dsd_config.region} --yes"

        try:
            # Note: if user can't create a project the returncode will be 6, not 1.
            #   This may affect whether a CompletedProcess is returned, or an Exception
            # is raised.
            # Also, create command outputs project id to stdout if known, all other
            # output goes to stderr.
            plugin_utils.run_slow_command(cmd)
        except subprocess.CalledProcessError as e:
            error_msg = upsun_msgs.unknown_create_error(e)
            raise DSDCommandError(error_msg)

        # Fix bug ignoring .upsun/local on Windows.
        if (msg_fixed := upsun_utils.fix_git_exclude_bug()):
            plugin_utils.write_output(msg_fixed)

    def _modify_settings(self):
        """Add upsun-specific settings.

        This settings block is currently the same for all users. The ALLOWED_HOSTS
        setting should be customized.
        """
        template_path = self.templates_path / "settings.py"
        plugin_utils.modify_settings_file(template_path)

    def _add_platform_app_yaml(self):
        """Add a .platform.app.yaml file."""

        # Build contents from template.
        if dsd_config.pkg_manager == "poetry":
            template_path = "poetry.platform.app.yaml"
        elif dsd_config.pkg_manager == "pipenv":
            template_path = "pipenv.platform.app.yaml"
        else:
            template_path = "platform.app.yaml"
        template_path = self.templates_path / template_path

        context = {
            "project_name": dsd_config.local_project_name,
            "deployed_project_name": self.deployed_project_name,
        }

        contents = plugin_utils.get_template_string(template_path, context)

        # Write file to project.
        path = dsd_config.project_root / ".platform.app.yaml"
        plugin_utils.add_file(path, contents)

    def _add_requirements(self):
        """Add requirements for Upsun."""
        requirements = ["platformshconfig", "gunicorn", "psycopg2"]
        plugin_utils.add_packages(requirements)

    def _add_platform_dir(self):
        """Add a .platform directory, if it doesn't already exist."""
        self.platform_dir_path = dsd_config.project_root / ".platform"
        plugin_utils.add_dir(self.platform_dir_path)

    def _add_services_yaml(self):
        """Add the .platform/services.yaml file."""

        template_path = self.templates_path / "services.yaml"
        contents = plugin_utils.get_template_string(template_path, context=None)

        path = self.platform_dir_path / "services.yaml"
        plugin_utils.add_file(path, contents)

    def _settings_env_var(self):
        """Set the DJANGO_SETTINGS_MODULE env var, if needed."""
        # This is primarily for Wagtail projects, as signified by a settings/production.py file.
        if dsd_config.settings_path.parts[-2:] == ("settings", "production.py"):
            plugin_utils.write_output(
                "  Setting DJANGO_SETTINGS_MODULE environment variable..."
            )

            # Need form mysite.settings.production
            dotted_settings_path = ".".join(
                dsd_config.settings_path.parts[-3:]
            ).removesuffix(".py")

            cmd = f"upsun variable:create --level environment --environment main --name DJANGO_SETTINGS_MODULE --value {dotted_settings_path} --no-interaction --visible-build true --prefix env"
            output = plugin_utils.run_quick_command(cmd)
            plugin_utils.write_output(output)

    def _conclude_automate_all(self):
        """Finish automating the push to Upsun.

        - Commit all changes.
        - Call `upsun push`.
        - Open project.
        """
        # Making this check here lets deploy() be cleaner.
        if not dsd_config.automate_all:
            return

        plugin_utils.commit_changes()

        # Push project.
        plugin_utils.write_output("  Pushing to Upsun...")

        # Pause to make sure project that was just created can be used.
        plugin_utils.write_output(
            "    Pausing 10s to make sure project is ready to use..."
        )
        time.sleep(10)

        # Use run_slow_command(), to stream output as it runs.
        cmd = "upsun push --yes"
        plugin_utils.run_slow_command(cmd)

        # Open project.
        plugin_utils.write_output("  Opening deployed app in a new browser tab...")
        cmd = "upsun url --yes"
        output = plugin_utils.run_quick_command(cmd)
        plugin_utils.write_output(output)

        # Get url of deployed project.
        #   This can be done with an re, but there's one line of output with
        #   a url, so finding that line is simpler.
        # DEV: Move this to a utility, and write a test against standard Upsun
        # output.
        self.deployed_url = ""
        for line in output.stdout.decode().split("\n"):
            if "https" in line:
                self.deployed_url = line.strip()

    def _show_success_message(self):
        """After a successful run, show a message about what to do next."""

        # DEV:
        # - Mention that this script should not need to be run again unless creating
        #   a new deployment.
        # - Describe ongoing approach of commit, push, migrate. Lots to consider
        #   when doing this on production app with users, make sure you learn.

        if dsd_config.automate_all:
            msg = upsun_msgs.success_msg_automate_all(self.deployed_url)
            plugin_utils.write_output(msg)
        else:
            msg = upsun_msgs.success_msg(dsd_config.log_output)
            plugin_utils.write_output(msg)

    # --- Helper methods for methods called from deploy.py ---

    def _check_upsun_settings(self):
        """Check to see if an Upsun settings block already exists."""
        start_line = "# Upsun settings."
        plugin_utils.check_settings(
            "Upsun",
            start_line,
            upsun_msgs.upsun_settings_found,
            upsun_msgs.cant_overwrite_settings,
        )

    def _validate_cli(self):
        """Make sure the Upsun CLI is installed, and user is authenticated."""
        cmd = "upsun --version"

        # This generates a FileNotFoundError on macOS and Linux if the CLI is not installed.
        try:
            output_obj = plugin_utils.run_quick_command(cmd)
        except FileNotFoundError:
            raise DSDCommandError(upsun_msgs.cli_not_installed)

        plugin_utils.log_info(output_obj)
        if sys.platform == "win32":
            stderr = output_obj.stderr.decode(encoding="utf-8")
            if "'upsun' is not recognized as an internal or external command" in stderr:
                raise DSDCommandError(upsun_msgs.cli_not_installed)

        # Check that the user is authenticated.
        cmd = "upsun auth:info --no-interaction"
        output_obj = plugin_utils.run_quick_command(cmd)

        if "Authentication is required." in output_obj.stderr.decode():
            raise DSDCommandError(upsun_msgs.cli_logged_out)

    def _get_upsun_project_name(self):
        """Get the deployed project name.

        If using automate_all, we'll set this. Otherwise, we're looking for the name
        that was given in the `upsun create` command.
        - Try to get this from `project:info`.
        - If can't get project name:
          - Exit with warning, and inform user of --deployed-project-name
            flag to override this error.

        Retuns:
            str: The deployed project name.
        Raises:
            DSDCommandError: If deployed project name can't be found.
        """
        # If we're creating the project, we'll just use the startproject name.
        if dsd_config.automate_all:
            return dsd_config.local_project_name

        # Use the provided name if --deployed-project-name specified.
        if dsd_config.deployed_project_name:
            return dsd_config.deployed_project_name

        # Use --yes flag to avoid interactive prompt hanging in background
        #   if the user is not currently logged in to the CLI.
        cmd = "upsun project:info --yes --format csv"
        output_obj = plugin_utils.run_quick_command(cmd)
        output_str = output_obj.stdout.decode()

        # Log cmd, but don't log the output of `project:info`. It contains identifying
        # information about the user and project, including client_ssh_key.
        plugin_utils.log_info(cmd)

        # If there's no stdout, the user is probably logged out, hasn't called
        #   create, or doesn't have the CLI installed.
        # Also, I've seen both ProjectNotFoundException and RootNotFoundException
        #   raised when no project has been created.
        if not output_str:
            output_str = output_obj.stderr.decode()
            if "LoginRequiredException" in output_str:
                raise DSDCommandError(upsun_msgs.login_required)
            elif "ProjectNotFoundException" in output_str:
                raise DSDCommandError(upsun_msgs.no_project_name)
            elif "RootNotFoundException" in output_str:
                raise DSDCommandError(upsun_msgs.no_project_name)
            else:
                error_msg = upsun_msgs.unknown_error
                error_msg += upsun_msgs.cli_not_installed
                raise DSDCommandError(error_msg)

        # Pull deployed project name from output.
        lines = output_str.splitlines()
        title_line = [line for line in lines if "title," in line][0]
        # Assume first project is one to use.
        project_name = title_line.split(",")[1].strip()
        project_name = upsun_utils.get_project_name(output_str)

        # Project names can only have lowercase alphanumeric characters.
        # See: https://github.com/ehmatthes/django-simple-deploy/issues/323
        if " " in project_name:
            project_name = project_name.replace(" ", "_").lower()
        if project_name:
            return project_name

        # Couldn't find a project name. Warn user, and tell them about override flag.
        raise DSDCommandError(upsun_msgs.no_project_name)

    def _get_org_name(self):
        """Get the organization name associated with the user's Upsun account.

        This is needed for creating a project using automate_all.
        Confirm that it's okay to use this org.

        Note: In the csv output, Upsun refers to the alphanumeric ID as a Name,
        and the user-provided name as a Label. This gets confusing. :/

        Returns:
            str: org name (id)
            None: if not using automate-all
        Raises:
            DSDCommandError:
            - if org name found, but not confirmed.
            - if org name not found
        """
        if not dsd_config.automate_all:
            return

        cmd = "upsun organization:list --yes --format csv"
        output_obj = plugin_utils.run_quick_command(cmd)
        output_str = output_obj.stdout.decode()
        plugin_utils.log_info(output_str)

        org_ids, org_names = upsun_utils.get_org_ids_names(output_str)

        if not org_names:
            raise DSDCommandError(upsun_msgs.org_not_found)

        if len(org_names) == 1:
            # Get permission to use this org.
            org_name = org_names[0]
            if self._confirm_use_org(org_name):
                # Return the corresponding ID, not the name.
                return org_ids[0]

        # Show all org names, ask user to make selection.
        prompt = "\n*** Found multiple organizations on Upsun. ***\n"
        for index, name in enumerate(org_names):
            prompt += f"\n  {index}: {name}"
        prompt += "\n\nWhich organization would you like to use? "

        valid_choices = [i for i in range(len(org_names))]

        # Confirm selection, because we do *not* want to deploy using the wrong org.
        confirmed = False
        while not confirmed:
            selection = plugin_utils.get_numbered_choice(
                prompt, valid_choices, upsun_msgs.no_org_available
            )
            selected_org = org_names[selection]

            confirm_prompt = f"You have selected {selected_org}."
            confirm_prompt += " Is that correct?"
            confirmed = plugin_utils.get_confirmation(confirm_prompt)

            # Return corresponding ID, not the name.
            return org_ids[selection]

    def _confirm_use_org(self, org_name):
        """Confirm that it's okay to use the org that was found.

        Returns:
            True: if confirmed
            DSDCommandError: if not confirmed
        """

        dsd_config.stdout.write(upsun_msgs.confirm_use_org(org_name))
        confirmed = plugin_utils.get_confirmation(skip_logging=True)

        if confirmed:
            dsd_config.stdout.write("  Okay, continuing with deployment.")
            return True
        else:
            # Exit, with a message that configuration is still an option.
            msg = upsun_msgs.cancel_upsun
            msg += upsun_msgs.may_configure
            raise DSDCommandError(msg)



"""A collection of messages used in platform_deployer.py."""

# For conventions, see documentation in deploy_messages.py

from textwrap import dedent

from django.conf import settings


confirm_automate_all = """
The --automate-all flag means the deploy command will:
- Run `upsun create` for you, to create an empty Upsun project.
  - This will create a project in the us-3.platform.sh region. If you wish
    to use a different region, cancel this operation and use the --region flag.
  - You can see a list of all regions at:
    https://docs.upsun.com/development/regions.html#region-location
- Commit all changes to your project that are necessary for deployment.
- Push these changes to Upsun.
- Open your deployed project in a new browser tab.
"""

cancel_upsun = """
Okay, cancelling Upsun deployment.
"""

cli_not_installed = """
In order to deploy to Upsun, you need to install the Upsun CLI.
  See here: https://docs.upsun.com/administration/cli.html
After installing the CLI, you can run the deploy command again.
"""

cli_logged_out = """
You are currently logged out of the Upsun CLI. Please log in,
  and then run the deploy command again.
You can log in from  the command line:
  $ upsun login
"""

upsun_settings_found = """
There is already an Upsun-specific settings block in settings.py. Is it okay to
overwrite this block, and everything that follows in settings.py?
"""

cant_overwrite_settings = """
In order to configure the project for deployment, we need to write an Upsun-specific
settings block. Please remove the current Upsun-specific settings, and then run
the deploy command again.
"""

no_project_name = """
An Upsun project name could not be found.

The deploy command expects that you've already run `upsun create`, or
associated the local project with an existing project on Upsun.

If you haven't done so, run the `upsun create` command and then run
the deploy command again. You can override this warning by using
the `--deployed-project-name` flag to specify the name you want to use for the
project. This must match the name of your Upsun project.
"""

org_not_found = """
An Upsun organization name could not be found.

You may have created an Upsun account, but not created an organization.
The Upsun CLI requires an organization name when creating a new project.

Please visit the Upsun console and make sure you have created an organization.
You can also do this through the CLI using the `upsun organization:create` command.
For help, run `upsun help organization:create`.
"""

no_org_available = """
An Upsun org must be used to make a deployment. Please identify or create the org
you'd like to use, and then try again.
"""

login_required = """
You appear to be logged out of the Upsun CLI. Please run the 
command `upsun login`, and then run the deploy command again.

You may be able to override this error by passing the `--deployed-project-name`
flag.
"""

unknown_error = """
An unknown error has occurred. Do you have the Upsun CLI installed?
"""

may_configure = """
You may want to re-run the deploy command without the --automate-all flag.

You will have to create the Upsun project yourself, but django-simple-deploy
will do all the necessary configuration for deployment.
"""


# --- Dynamic strings ---
# These need to be generated in functions, to display information that's
#   determined as the script runs.


def confirm_use_org(org_name):
    """Confirm use of this org to create a new project."""

    msg = dedent(
        f"""
        --- The Upsun CLI requires an organization name when creating a new project. ---
        When using --automate-all, a project will be created on your behalf. The following
        organization was found: {org_name}

        This organization will be used to create a new project. If this is not okay,
        enter n to cancel this operation.
    """
    )

    return msg


def unknown_create_error(e):
    """Process a non-specific error when running `upsun create`
    while using automate_all. This is most likely an issue with the user
    not having permission to create a new project, for example because they
    are on a trial plan and have already created too many projects.
    """

    msg = dedent(
        f"""
        --- An error has occurred when trying to create a new Upsun project. ---

        While running `upsun create`, an error has occurred. You should check
        the Upsun console to see if a project was partially created.

        The error messages that Upsun provides, both through the CLI and
        the console, are not always specific enough to be helpful.

        The following output may help diagnose the error:
        ***** output of `upsun create` *****

        {e.stderr.decode()}

        ***** end output *****
    """
    )

    return msg


def success_msg(log_output=""):
    """Success message, for configuration-only run."""

    msg = dedent(
        f"""
        --- Your project is now configured for deployment on Upsun. ---

        To deploy your project, you will need to:
        - Commit the changes made in the configuration process.
            $ git status
            $ git add .
            $ git commit -am "Configured project for deployment."
        - Push your project to Upsun's servers:
            $ upsun push
        - Open your project:
            $ upsun url    
        - As you develop your project further:
            - Make local changes
            - Commit your local changes
            - Run `upsun push`
    """
    )

    if log_output:
        msg += dedent(
            f"""
        - You can find a full record of this configuration in the dsd_logs directory.
        """
        )

    return msg


def success_msg_automate_all(deployed_url):
    """Success message, when using --automate-all."""

    msg = dedent(
        f"""

        --- Your project should now be deployed on Upsun. ---

        It should have opened up in a new browser tab.
        - You can also visit your project at {deployed_url}

        If you make further changes and want to push them to Upsun,
        commit your changes and then run `upsun push`.

        Also, if you haven't already done so you should review the
        documentation for Python deployments on Upsun at:
        - https://fixed.docs.upsun.com/languages/python.html
        - This documentation will help you understand how to maintain
          your deployment.

    """
    )
    return msg



"""Config class for plugin information shared with core."""

from . import deploy_messages as upsun_msgs


class PluginConfig:
    """Class for managing attributes that need to be shared with core.

    This is similar to the class DSDConfig in core's dsd_config.py.

    This should future-proof plugins somewhat, in that if more information needs
    to be shared back to core, it can be added here without breaking changes to the
    core-plugin interface.

    Get plugin-specific attributes required by core.

    Required:
    - automate_all_supported
    - platform_name
    Optional:
    - confirm_automate_all_msg (required if automate_all_supported is True)
    """

    def __init__(self):
        self.automate_all_supported = True
        self.confirm_automate_all_msg = upsun_msgs.confirm_automate_all
        self.platform_name = "Upsun"
